@@ -2,7 +2,7 @@
 # -*- coding: utf-8 -*-
 #
 # Copyright 2005 Lars Wirzenius (liw@iki.fi)
-# Copyright © 2011-2013 Andreas Beckmann (anbe@debian.org)
+# Copyright © 2011-2019 Andreas Beckmann (anbe@debian.org)
 #
 # This program is free software; you can redistribute it and/or modify it
 # under the terms of the GNU General Public License as published by the
@@ -15,15 +15,14 @@
 # Public License for more details.
 #
 # You should have received a copy of the GNU General Public License along with
-# this program; if not, write to the Free Software Foundation, Inc.,
-# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
+# this program. If not, see <https://www.gnu.org/licenses/>
 
 
 """Distributed piuparts processing, slave program
 
 Lars Wirzenius <liw@iki.fi>
 """
-
+from __future__ import print_function
 
 import os
 import sys
@@ -34,8 +33,8 @@ from signal import alarm, signal, SIGALRM, SIGINT, SIGKILL, SIGHUP
 import subprocess
 import fcntl
 import random
-import ConfigParser
 import apt_pkg
+import pipes
 
 import piupartslib.conf
 import piupartslib.packagesdb
@@ -46,12 +45,13 @@ apt_pkg.init_system()
 
 CONFIG_FILE = "/etc/piuparts/piuparts.conf"
 DISTRO_CONFIG_FILE = "/etc/piuparts/distros.conf"
-MAX_WAIT_TEST_RUN = 45*60
+MAX_WAIT_TEST_RUN = 90 * 60
 
 interrupted = False
 old_sigint_handler = None
 got_sighup = False
 
+
 def setup_logging(log_level, log_file_name):
     logger = logging.getLogger()
     logger.setLevel(log_level)
@@ -73,52 +73,58 @@ class Config(piupartslib.conf.Config):
     def __init__(self, section="slave", defaults_section=None):
         self.section = section
         piupartslib.conf.Config.__init__(self, section,
-            {
-                "sections": "slave",
-                "idle-sleep": 300,
-                "max-tgz-age": 2592000,
-                "min-tgz-retry-delay": 21600,
-                "master-host": None,
-                "master-user": None,
-                "master-command": None,
-                "proxy": None,
-                "mirror": None,
-                "piuparts-command": "sudo piuparts",
-                "piuparts-flags": "",
-                "tmpdir": None,
-                "distro": None,
-                "area": None,
-                "components": None,
-                "chroot-tgz": None,
-                "upgrade-test-distros": None,
-                "basetgz-directory": ".",
-                "max-reserved": 1,
-                "debug": "no",
-                "keep-sources-list": "no",
-                "arch": None,
-                "precedence": "1",
-                "slave-load-max": None,
-            },
-            defaults_section=defaults_section)
+                                         {
+                                         "sections": "slave",
+                                         "basetgz-sections": "",
+                                         "idle-sleep": 300,
+                                         "max-tgz-age": 2592000,
+                                         "min-tgz-retry-delay": 21600,
+                                         "master-host": None,
+                                         "master-user": None,
+                                         "master-command": None,
+                                         "proxy": None,
+                                         "mirror": None,
+                                         "setarch": None,
+                                         "piuparts-command": "sudo piuparts",
+                                         "piuparts-flags": "",
+                                         "tmpdir": None,
+                                         "distro": None,
+                                         "area": None,
+                                         "components": None,
+                                         "chroot-tgz": None,
+                                         "upgrade-test-distros": None,
+                                         "basetgz-directory": ".",
+                                         "chroot-meta-auto": None,
+                                         "max-reserved": 1,
+                                         "debug": "no",
+                                         "keep-sources-list": "no",
+                                         "arch": None,
+                                         "precedence": "1",
+                                         "slave-load-max": None,
+                                         },
+                                         defaults_section=defaults_section)
 
 
 class Alarm(Exception):
     pass
 
+
 def alarm_handler(signum, frame):
     raise Alarm
 
+
 def sigint_handler(signum, frame):
     global interrupted
     interrupted = True
-    print '\nSlave interrupted by the user, waiting for the current test to finish.'
-    print 'Press Ctrl-C again to abort now.'
+    print('\nSlave interrupted by the user, waiting for the current test to finish.')
+    print('Press Ctrl-C again to abort now.')
     signal(SIGINT, old_sigint_handler)
 
+
 def sighup_handler(signum, frame):
     global got_sighup
     got_sighup = True
-    print 'SIGHUP: Will flush finished logs.'
+    print('SIGHUP: Will flush finished logs.')
 
 
 class MasterIsBusy(Exception):
@@ -254,12 +260,11 @@ class Slave:
         package, rest = basename.split("_", 1)
         version = rest[:-len(".log")]
         self._writeline(pass_or_fail, package, version)
-        f = file(filename, "r")
-        for line in f:
-            if line.endswith("\n"):
-                line = line[:-1]
-            self._writeline(" " + line)
-        f.close()
+        with open(filename, "r") as f:
+            for line in f:
+                if line.endswith("\n"):
+                    line = line[:-1]
+                self._writeline(" " + line)
         self._writeline(".")
         line = self._readline()
         if line != "ok\n":
@@ -315,7 +320,7 @@ class Slave:
             raise MasterNotOK()
 
     def _reserved_filename(self, name, version):
-        return os.path.join("reserved",  "%s_%s.log" % (name, version))
+        return os.path.join("reserved", "%s_%s.log" % (name, version))
 
     def remember_reservation(self, name, version):
         create_file(self._reserved_filename(name, version), "")
@@ -341,10 +346,11 @@ class Section:
         self._config = Config(section=section, defaults_section="global")
         self._config.read(CONFIG_FILE)
         self._distro_config = piupartslib.conf.DistroConfig(
-                DISTRO_CONFIG_FILE, self._config["mirror"])
+            DISTRO_CONFIG_FILE, self._config["mirror"])
         self._error_wait_until = 0
         self._idle_wait_until = 0
         self._recycle_wait_until = 0
+        self._tarball_wait_until = 0
         self._slave_directory = os.path.abspath(section)
         if not os.path.exists(self._slave_directory):
             os.makedirs(self._slave_directory)
@@ -363,7 +369,6 @@ class Section:
 
         self._slave = slave or Slave()
 
-
     def _throttle_if_overloaded(self):
         global interrupted
         if interrupted or got_sighup:
@@ -392,7 +397,6 @@ class Section:
             if secs < 300:
                 secs += random.randrange(30, 90)
 
-
     def _connect_to_master(self, recycle=False):
         self._slave.set_master_host(self._config["master-host"])
         self._slave.set_master_user(self._config["master-user"])
@@ -402,35 +406,42 @@ class Section:
         if recycle:
             self._slave.enable_recycling()
 
-
     def _get_tarball(self):
         basetgz = self._config["chroot-tgz"] or \
-                self._distro_config.get_basetgz(self._config.get_start_distro(),
-                                                self._config.get_arch())
+            self._distro_config.get_basetgz(self._config.get_start_distro(),
+                                            self._config.get_arch())
         return os.path.join(self._config["basetgz-directory"], basetgz)
 
     def _check_tarball(self):
+        if int(self._config["max-tgz-age"]) < 0:
+            return
+
         oldcwd = os.getcwd()
         os.chdir(self._slave_directory)
 
         tgz = self._get_tarball()
         max_tgz_age = int(self._config["max-tgz-age"])
         min_tgz_retry_delay = int(self._config["min-tgz-retry-delay"])
+        ttl = 0
         needs_update = not os.path.exists(tgz)
         if not needs_update and max_tgz_age > 0:
             # tgz exists and age is limited, so check age
             now = time.time()
             age = now - os.path.getmtime(tgz)
+            ttl = max_tgz_age - age
             logging.info("Check-replace %s: age=%d vs. max=%d" % (tgz, age, max_tgz_age))
-            if age > max_tgz_age:
+            if ttl < 0:
                 if os.path.exists(tgz + ".log"):
                     age = now - os.path.getmtime(tgz + ".log")
+                ttl = min_tgz_retry_delay - age
                 logging.info("Limit-replace %s: last-retry=%d vs. min=%d" % (tgz, age, min_tgz_retry_delay))
-                if age > min_tgz_retry_delay:
+                if ttl < 0:
                     needs_update = True
                     logging.info("%s too old.  Forcing re-creation" % tgz)
         if needs_update:
             create_chroot(self._config, tgz, self._config.get_start_distro())
+            ttl = min_tgz_retry_delay
+        self._tarball_wait_until = time.time() + ttl
 
         os.chdir(oldcwd)
 
@@ -453,13 +464,34 @@ class Section:
             return max(self._error_wait_until, self._recycle_wait_until)
         return max(self._error_wait_until, self._idle_wait_until)
 
-
     def run(self, do_processing=True, recycle=False):
         if time.time() < self.sleep_until(recycle=recycle):
             return 0
 
         self._throttle_if_overloaded()
 
+        self._config = Config(section=self._config.section, defaults_section="global")
+        try:
+            self._config.read(CONFIG_FILE)
+        except MissingSection:
+            logging.info("unknown section " + self._config.section)
+            self._error_wait_until = time.time() + 3600
+            return 0
+        self._distro_config = piupartslib.conf.DistroConfig(
+                DISTRO_CONFIG_FILE, self._config["mirror"])
+
+        if interrupted or got_sighup:
+            do_processing = False
+
+        if do_processing and time.time() > self._tarball_wait_until:
+            self._check_tarball()
+
+        if self._config.get_distro() == "None":
+            # section is for tarball creation only
+            self._idle_wait_until = self._tarball_wait_until + 60
+            self._recycle_wait_until = self._tarball_wait_until + 3600
+            return 0
+
         if interrupted or got_sighup:
             do_processing = False
 
@@ -472,19 +504,9 @@ class Section:
             action = "Recycling"
         if not do_processing:
             action = "Flushing"
-        logging.info("%s section %s (precedence=%d)" \
+        logging.info("%s section %s (precedence=%d)"
                      % (action, self._config.section, self.precedence()))
 
-        self._config = Config(section=self._config.section, defaults_section="global")
-        try:
-            self._config.read(CONFIG_FILE)
-        except MissingSection:
-            logging.info("unknown")
-            self._error_wait_until = time.time() + 3600
-            return 0
-        self._distro_config = piupartslib.conf.DistroConfig(
-                DISTRO_CONFIG_FILE, self._config["mirror"])
-
         if int(self._config["max-reserved"]) == 0:
             logging.info("disabled")
             self._error_wait_until = time.time() + 12 * 3600
@@ -525,7 +547,6 @@ class Section:
                 os.chdir(oldcwd)
         return 0
 
-
     def _talk_to_master(self, fetch=False, unreserve=False, recycle=False):
         flush = self._count_submittable_logs() > 0
         fetch = fetch and not self._slave.get_reserved()
@@ -542,7 +563,7 @@ class Section:
         except MasterCantRecycle:
             logging.error("master has nothing to recycle")
             self._recycle_wait_until = max(time.time(), self._idle_wait_until) + 3600
-        except (MasterDidNotGreet, MasterIsCrazy, MasterCommunicationFailed):
+        except (MasterDidNotGreet, MasterIsCrazy, MasterCommunicationFailed, MasterNotOK):
             logging.error("connection to master failed")
             self._error_wait_until = time.time() + 900
             self._slave.close()
@@ -589,8 +610,8 @@ class Section:
                 return True
         return False
 
-
     def _process(self):
+        global interrupted
         self._slave.close()
 
         packagenames = set([x[0] for x in self._slave.get_reserved()])
@@ -600,8 +621,8 @@ class Section:
                 try:
                     pf = piupartslib.packagesdb.PackagesFile()
                     pf.load_packages_urls(
-                            self._distro_config.get_packages_urls(
-                                distro,
+                        self._distro_config.get_packages_urls(
+                            distro,
                                 self._config.get_area(),
                                 self._config.get_arch()),
                             packagenames)
@@ -610,12 +631,23 @@ class Section:
                     logging.error("failed to fetch packages file for %s" % distro)
                     self._error_wait_until = time.time() + 900
                     return 0
+                except KeyboardInterrupt:
+                    interrupted = True
         del packagenames
 
         test_count = 0
         self._check_tarball()
         if not os.path.exists(self._get_tarball()):
             self._error_wait_until = time.time() + 300
+        if self._config["chroot-meta-auto"]:
+            if os.path.exists(self._config["chroot-meta-auto"]):
+                try:
+                    age = time.time() - os.path.getmtime(self._config["chroot-meta-auto"])
+                    if age > 6 * 3600:
+                        os.unlink(self._config["chroot-meta-auto"])
+                        logging.info("Deleting old %s" % self._config["chroot-meta-auto"])
+                except OSError:
+                    pass
         for package_name, version in self._slave.get_reserved():
             self._throttle_if_overloaded()
             if interrupted or got_sighup:
@@ -629,7 +661,6 @@ class Section:
         self._talk_to_master(unreserve=interrupted)
         return test_count
 
-
     def _test_package(self, pname, pvers, packages_files):
         global old_sigint_handler
         old_sigint_handler = signal(SIGINT, sigint_handler)
@@ -639,13 +670,17 @@ class Section:
         output_name = log_name(pname, pvers)
         logging.debug("Opening log file %s" % output_name)
         new_name = os.path.join("new", output_name)
-        output = file(new_name, "we")
+        output = open(new_name, "we")
         output.write(time.strftime("Start: %Y-%m-%d %H:%M:%S %Z\n",
                                    time.gmtime()))
 
         distupgrade = len(self._config.get_distros()) > 1
 
-        command = self._config["piuparts-command"].split()
+        command = []
+        if self._config["setarch"]:
+            command.append("setarch")
+            command.extend(self._config["setarch"].split())
+        command.extend(self._config["piuparts-command"].split())
         if self._config["piuparts-flags"]:
             command.extend(self._config["piuparts-flags"].split())
         if "http_proxy" in os.environ:
@@ -667,6 +702,11 @@ class Section:
                 command.extend(["-d", distro])
         if self._config["keep-sources-list"] in ["yes", "true"]:
             command.append("--keep-sources-list")
+        if distupgrade and self._config["chroot-meta-auto"]:
+            if not os.path.exists(self._config["chroot-meta-auto"]):
+                command.extend(["-S", self._config["chroot-meta-auto"]])
+            else:
+                command.extend(["-B", self._config["chroot-meta-auto"]])
         command.extend(["--apt", "%s=%s" % (pname, pvers)])
 
         subdir = "fail"
@@ -732,17 +772,27 @@ class Section:
             subdir = "untestable"
 
         if ret == 0:
-            output.write("Executing: %s\n" % " ".join(quote_spaces(command)))
+            output.write("Executing: %s\n" % command2string(command))
             ret, f = run_test_with_timeout(command, MAX_WAIT_TEST_RUN)
             if not f or f[-1] != '\n':
                 f += '\n'
-            output.write(f)
+            output.write(f.replace('\033', '[ESC]'))
             lastline = f.split('\n')[-2]
             if ret < 0:
                 output.write(" *** Process KILLED - exceed maximum run time ***\n")
             elif not "piuparts run ends" in lastline:
                 ret += 1024
-                output.write(" *** PIUPARTS OUTPUT INCOMPLETE ***\n");
+                output.write(" *** PIUPARTS OUTPUT INCOMPLETE ***\n")
+            elif distupgrade and self._config["chroot-meta-auto"]:
+                try:
+                    if "History of available packages does not match - reference chroot may be outdated" in f:
+                        os.unlink(self._config["chroot-meta-auto"])
+                        logging.info("Deleting outdated %s" % self._config["chroot-meta-auto"])
+                    elif "Initial package selections do not match - ignoring loaded reference chroot state" in f:
+                        os.unlink(self._config["chroot-meta-auto"])
+                        logging.info("Deleting mismatching %s" % self._config["chroot-meta-auto"])
+                except OSError:
+                    pass
 
         output.write("\n")
         output.write("ret=%d\n" % ret)
@@ -760,8 +810,9 @@ def log_name(package, version):
     return "%s_%s.log" % (package, version)
 
 
-def quote_spaces(vlist):
-    return ["'%s'" % x if ' ' in x else x for x in vlist]
+def command2string(command):
+    """Quote s.t. copy+paste from the logfile gives a runnable command in the shell."""
+    return " ".join([pipes.quote(arg) for arg in command])
 
 
 def run_test_with_timeout(cmd, maxwait, kill_all=True):
@@ -770,11 +821,11 @@ def run_test_with_timeout(cmd, maxwait, kill_all=True):
         pids = [p.pid]
         if kill_all:
             ps = subprocess.Popen(["ps", "--no-headers", "-o", "pid", "--ppid", "%d" % p.pid],
-                                  stdout = subprocess.PIPE)
+                                  stdout=subprocess.PIPE)
             stdout, stderr = ps.communicate()
             pids.extend([int(pid) for pid in stdout.split()])
         if p.poll() is None:
-            print 'Sending SIGINT...'
+            print('Sending SIGINT...')
             try:
                 os.killpg(os.getpgid(p.pid), SIGINT)
             except OSError:
@@ -785,7 +836,7 @@ def run_test_with_timeout(cmd, maxwait, kill_all=True):
                 if p.poll() is not None:
                     break
         if p.poll() is None:
-            print 'Sending SIGTERM...'
+            print('Sending SIGTERM...')
             p.terminate()
             # piuparts has 5 seconds to clean up after SIGTERM
             for i in range(10):
@@ -793,17 +844,17 @@ def run_test_with_timeout(cmd, maxwait, kill_all=True):
                 if p.poll() is not None:
                     break
         if p.poll() is None:
-            print 'Sending SIGKILL...'
+            print('Sending SIGKILL...')
             p.kill()
         for pid in pids:
             if pid > 0:
                 try:
                     os.kill(pid, SIGKILL)
-                    print "Killed %d" % pid
+                    print("Killed %d" % pid)
                 except OSError:
                     pass
 
-    logging.debug("Executing: %s" % " ".join(quote_spaces(cmd)))
+    logging.debug("Executing: %s" % command2string(cmd))
 
     stdout = ""
     p = subprocess.Popen(cmd, preexec_fn=os.setpgrp,
@@ -818,11 +869,11 @@ def run_test_with_timeout(cmd, maxwait, kill_all=True):
         terminate_subprocess(p, kill_all)
         return -1, stdout
     except KeyboardInterrupt:
-        print '\nSlave interrupted by the user, cleaning up...'
+        print('\nSlave interrupted by the user, cleaning up...')
         try:
             terminate_subprocess(p, kill_all)
         except KeyboardInterrupt:
-            print '\nTerminating piuparts was interrupted... manual cleanup still neccessary.'
+            print('\nTerminating piuparts was interrupted... manual cleanup still neccessary.')
             raise
         raise
 
@@ -834,7 +885,11 @@ def run_test_with_timeout(cmd, maxwait, kill_all=True):
 
 
 def create_chroot(config, tarball, distro):
-    command = config["piuparts-command"].split()
+    command = []
+    if config["setarch"]:
+        command.append("setarch")
+        command.extend(config["setarch"].split())
+    command.extend(config["piuparts-command"].split())
     if config["piuparts-flags"]:
         command.extend(config["piuparts-flags"].split())
     if "http_proxy" in os.environ:
@@ -849,7 +904,8 @@ def create_chroot(config, tarball, distro):
     command.extend(["--arch", config.get_arch()])
     command.extend(["-d", distro])
     command.extend(["-s", tarball + ".new"])
-    command.extend(["--apt", "dpkg"])
+    command.extend(['--no-install-purge-test', '--no-upgrade-test'])
+    command.extend(["--apt", "TARBALL"])  # dummy package name
 
     output_name = tarball + ".log"
     with open(output_name, "we") as output:
@@ -861,25 +917,29 @@ def create_chroot(config, tarball, distro):
             logging.info("Creating new tarball %s" % tarball)
             output.write(time.strftime("Start: %Y-%m-%d %H:%M:%S %Z\n\n",
                                        time.gmtime()))
-            output.write("Executing: " + " ".join(quote_spaces(command)) + "\n\n")
-            logging.debug("Executing: " + " ".join(quote_spaces(command)))
-            p = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
-            for line in p.stdout:
-                output.write(line)
-                logging.debug(">> " + line.rstrip())
-            p.wait()
-            output.write(time.strftime("\nEnd: %Y-%m-%d %H:%M:%S %Z\n",
-                                       time.gmtime()))
-            if os.path.exists(tarball + ".new"):
-                os.rename(tarball + ".new", tarball)
-            else:
-                logging.error("Tarball creation failed, see %s" % output_name)
+            output.write("Executing: " + command2string(command) + "\n\n")
+            logging.debug("Executing: " + command2string(command))
+            try:
+                p = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
+                for line in p.stdout:
+                    output.write(line)
+                    logging.debug(">> " + line.rstrip())
+                p.wait()
+                output.write(time.strftime("\nEnd: %Y-%m-%d %H:%M:%S %Z\n",
+                                           time.gmtime()))
+                if os.path.exists(tarball + ".new"):
+                    os.rename(tarball + ".new", tarball)
+                else:
+                    logging.error("Tarball creation failed, see %s" % output_name)
+            except IOError:
+                output.write(time.strftime("\nFAIL: %Y-%m-%d %H:%M:%S %Z\n",
+                                           time.gmtime()))
+                logging.error("Tarball creation failed with IOError")
 
 
 def create_file(filename, contents):
-    f = file(filename, "w")
-    f.write(contents)
-    f.close()
+    with open(filename, "w") as f:
+        f.write(contents)
 
 
 def main():
@@ -899,6 +959,7 @@ def main():
         section_names = sys.argv[1:]
     else:
         section_names = global_config["sections"].split()
+        section_names += global_config["basetgz-sections"].split()
 
     persistent_connection = Slave()
     sections = []
@@ -913,6 +974,10 @@ def main():
         logging.error("no sections found")
         return
 
+    # flush logs from previous run
+    for section in sections:
+        section.run(do_processing=False)
+
     while True:
         global got_sighup
         test_count = 0
@@ -951,8 +1016,8 @@ if __name__ == "__main__":
     try:
         main()
     except KeyboardInterrupt:
-        print ''
-        print 'Slave interrupted by the user, exiting...'
+        print('')
+        print('Slave interrupted by the user, exiting...')
         sys.exit(1)
 
 # vi:set et ts=4 sw=4 :
