@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 
 # Copyright 2005 Lars Wirzenius (liw@iki.fi)
-# Copyright © 2011-2013 Andreas Beckmann (anbe@debian.org)
+# Copyright © 2011-2018 Andreas Beckmann (anbe@debian.org)
 #
 # This program is free software; you can redistribute it and/or modify it
 # under the terms of the GNU General Public License as published by the
@@ -14,8 +14,7 @@
 # Public License for more details.
 #
 # You should have received a copy of the GNU General Public License along with
-# this program; if not, write to the Free Software Foundation, Inc.,
-# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
+# this program. If not, see <https://www.gnu.org/licenses/>
 
 
 """Packages database for distributed piuparts processing
@@ -30,6 +29,7 @@ Lars Wirzenius <liw@iki.fi>
 import logging
 import os
 import random
+import stat
 import tempfile
 import time
 import UserDict
@@ -43,7 +43,7 @@ apt_pkg.init_system()
 
 def rfc822_like_header_parse(input):
     headers = []
-    while 1:
+    while True:
         line = input.readline()
         if not line or line in ["\r\n", "\n"]:
             break
@@ -53,6 +53,7 @@ def rfc822_like_header_parse(input):
             headers.append(line)
     return headers
 
+
 class Package(UserDict.UserDict):
 
     def __init__(self, headers):
@@ -63,9 +64,47 @@ class Package(UserDict.UserDict):
             self[name.strip()] = value.strip()
         self._parsed_deps = {}
         self._parsed_alt_deps = {}
-        self._rrdep_count = None
-        self._block_count = None
-        self._waiting_count = None
+        self.rrdep_cnt = None
+        self.block_cnt = None
+        self.waiting_cnt = None
+        self.rdep_chain_len = None
+
+    def name(self):
+        return self["Package"]
+
+    def version(self):
+        return self["Version"]
+
+    def source(self):
+        # Binary packages built from the source package with the same name
+        # and version don't have a Source header.
+        if "Source" in self:
+            # If source and binary version differ (e.g. for binNMUs), the
+            # source version is given as the second element in the "Source"
+            # entry. Strip off the optional source version.
+            return self["Source"].split(" ")[0]
+        return self["Package"]
+
+    def source_version(self):
+        # Binary packages built from the source package with the same name
+        # and version don't have a Source header.
+        if "Source" in self:
+            # If source and binary version differ (e.g. for binNMUs), the
+            # source version is given as the second element in the "Source"
+            # entry. Strip off the parentheses around the source version.
+            if " " in self["Source"]:
+                return self["Source"].split(" ")[1][1:-1]
+        return self["Version"]
+
+    def set_test_versions(self, tv):
+        self["TestVersions"] = tv
+
+    def test_versions(self):
+        """ToDo: for distupgrade tests test_versions() should return a list
+           of versions in all distros in the upgrade path"""
+        if "TestVersions" in self:
+            return self["TestVersions"]
+        return self.version()
 
     def _parse_dependencies(self, header_name):
         if header_name in self._parsed_deps:
@@ -120,37 +159,6 @@ class Package(UserDict.UserDict):
                 vlist += self._parse_dependencies(header)
         return vlist
 
-    def is_testable(self):
-        """Are we testable at all? Required aren't."""
-        return self.get("Priority", "") != "required"
-
-    def rrdep_count(self):
-        """Get the recursive dependency count, if it has been calculated"""
-        if self._rrdep_count == None:
-            raise Exception('Reverse dependency count has not been calculated')
-        return self._rrdep_count
-
-    def set_rrdep_count(self, val):
-        self._rrdep_count = val
-
-    def block_count(self):
-        """Get the number of packages blocked by this package"""
-        if self._block_count == None:
-            raise Exception('Block count has not been calculated')
-        return self._block_count
-
-    def set_block_count(self, val):
-        self._block_count = val
-
-    def waiting_count(self):
-        """Get the number of packages waiting for this package"""
-        if self._waiting_count == None:
-            raise Exception('Waiting count has not been calculated')
-        return self._waiting_count
-
-    def set_waiting_count(self, val):
-        self._waiting_count = val
-
     def dump(self, output_file):
         output_file.write("".join(self.headers))
 
@@ -159,15 +167,18 @@ class PackagesFile(UserDict.UserDict):
 
     def __init__(self):
         UserDict.UserDict.__init__(self)
+        self._urllist = []
 
-    def load_packages_urls(self, urls):
+    def load_packages_urls(self, urls, restrict_packages=None):
         for url in urls:
+            logging.debug("Opening %s.*" % url)
+            (url, stream) = piupartslib.open_packages_url(url)
             logging.debug("Fetching %s" % url)
-            stream = piupartslib.open_packages_url(url)
-            self._read_file(stream)
+            self._read_file(stream, restrict_packages=restrict_packages)
             stream.close()
+            self._urllist.append(url)
 
-    def _read_file(self, input):
+    def _read_file(self, input, restrict_packages=None):
         """Parse a Packages file and add its packages to us-the-dict"""
         while True:
             headers = rfc822_like_header_parse(input)
@@ -179,8 +190,15 @@ class PackagesFile(UserDict.UserDict):
                 if apt_pkg.version_compare(p["Version"], q["Version"]) <= 0:
                     # there is already a newer version
                     continue
+            if restrict_packages is not None:
+                if p["Package"] not in restrict_packages:
+                    # unwanted package
+                    continue
             self[p["Package"]] = p
 
+    def get_urls(self):
+        return self._urllist
+
 
 class LogDB:
 
@@ -212,9 +230,6 @@ class LogDB:
             if basename.endswith(".log"):
                 cache[os.path.join(dirname, basename)] = True
 
-    def open_file(self, pathname, mode):
-        return file(pathname, mode)
-
     def remove_file(self, pathname):
         os.remove(pathname)
 
@@ -222,7 +237,7 @@ class LogDB:
         return "%s_%s.log" % (package, version)
 
     def log_exists(self, package, subdirs):
-        log_name = self._log_name(package["Package"], package["Version"])
+        log_name = self._log_name(package.name(), package.test_versions())
         for subdir in subdirs:
             if self.exists(os.path.join(subdir, log_name)):
                 return True
@@ -238,7 +253,7 @@ class LogDB:
         # Let's make it follow the umask.
         umask = os.umask(0)
         os.umask(umask)
-        os.chmod(temp_name, 0666 & ~umask)
+        os.chmod(temp_name, 0o666 & ~umask)
 
         full_name = os.path.join(subdir, self._log_name(package, version))
         try:
@@ -256,6 +271,10 @@ class LogDB:
             self.remove_file(full_name)
         self._evict(full_name)
 
+    def stat(self, subdir, package, version):
+        full_name = os.path.join(subdir, self._log_name(package, version))
+        return os.stat(full_name)
+
 
 class LogfileExists(Exception):
 
@@ -265,13 +284,22 @@ class LogfileExists(Exception):
 
 class PackagesDB:
 
+    # these packages are uses as dependencies but are only available
+    # from foreign architectures
+    # HACK: this hardcoded list should be moved to some data file
+    _foreign_packages = {
+        "ia32-libs-i386": "i386",
+        "ia32-libs-gtk-i386": "i386",
+        "libnss-mdns-i386": "i386",
+    }
+
     # keep in sync with piuparts-report.py: emphasize_reason()
     # FIXME: can we reorder this list or remove entries without breaking the counts.txt for the plot?
     _states = [
         "successfully-tested",
         "failed-testing",
         "cannot-be-tested",
-        "essential-required",
+        "essential-required",  # obsolete
         "waiting-to-be-tested",
         "waiting-for-dependency-to-be-tested",
         "dependency-failed-testing",
@@ -281,15 +309,18 @@ class PackagesDB:
         "unknown",
         "unknown-preferred-alternative",  # obsolete
         "no-dependency-from-alternatives-exists",  # obsolete
-        #"does-not-exist",  # can only happen as query result for a dependency
+        "outdated",
+        # "foreign:*",  # can only happen as query result for a dependency
+        # "does-not-exist",  # can only happen as query result for a dependency
     ]
 
     _good_states = [
         "successfully-tested",
         "essential-required",
-    ]
+    ] + ["foreign:%s" % arch for arch in set(_foreign_packages.values())]
 
     _obsolete_states = [
+        "essential-required",
         "circular-dependency",
         "unknown-preferred-alternative",
         "no-dependency-from-alternatives-exists",
@@ -320,6 +351,7 @@ class PackagesDB:
         self._dependency_databases = []
         self._recycle_mode = False
         self._candidates_for_testing = None
+        self._rdeps = None
         self.set_subdirs(ok="pass", fail="fail", evil="untestable",
                          reserved="reserved", morefail=["bugged", "affected"],
                          recycle="recycle")
@@ -379,6 +411,22 @@ class PackagesDB:
         self._packages_files.append(pf)
         self._packages = None
 
+    def load_alternate_versions_from_packages_urls(self, urls):
+        # take version numbers (or None) from alternate URLs
+        pf2 = PackagesFile()
+        pf2.load_packages_urls(urls)
+        for package in self.get_all_packages():
+            if package.name() in pf2:
+                package.set_test_versions(pf2[package.name()].version())
+            else:
+                package.set_test_versions("None")
+
+    def get_urls(self):
+        urls = []
+        for pf in self._packages_files:
+            urls.extend(pf.get_urls())
+        return urls
+
     def set_dependency_databases(self, dependency_databases=[]):
         self._dependency_databases = list(dependency_databases)
 
@@ -425,18 +473,40 @@ class PackagesDB:
                     more += dep_pkg.dependencies()
         return circular
 
-    def _lookup_package_state(self, package):
+    def _is_successfully_tested(self, package):
+        # a pass/ log exists but no corresponding recycle/ log exists
+        if self._logdb.log_exists(package, [self._ok]):
+            if not (self._recycle_mode and self._logdb.log_exists(package, [self._recycle])):
+                return True
+        return False
+
+    def _lookup_package_state(self, package, use_cached_success, check_outdated):
+        if check_outdated:
+            # Check if dependency databases have a newer version of this package.
+            # Use the actual package versions, not the target versions.
+            curr_ver = package.version()
+            for db in self._dependency_databases:
+                dep_ver = db.get_version(package.name())
+                if dep_ver is not None and apt_pkg.version_compare(curr_ver, dep_ver) < 0:
+                    #logging.info("[%s] outdated: %s %s < %s @[%s]" % (self.prefix, package.name(), curr_ver, dep_ver, db.prefix))
+                    return "outdated";
         if self._recycle_mode and self._logdb.log_exists(package, [self._recycle]):
             return "unknown"
         if self._logdb.log_exists(package, [self._ok]):
-            return "successfully-tested"
+            success = True
+            if not use_cached_success:
+                # if a pass/ log exists but any dependency may be not
+                # trivially satisfiable do not skip dependency resolution
+                for dep in package.dependencies():
+                    if not self.get_package(dep, resolve_virtual=True):
+                        success = False
+                        break
+            if success:
+                return "successfully-tested"
         if self._logdb.log_exists(package, [self._fail] + self._morefail):
             return "failed-testing"
         if self._logdb.log_exists(package, [self._evil]):
             return "cannot-be-tested"
-        if not package.is_testable():
-            return "essential-required"
-
         return "unknown"
 
     def _compute_package_state(self, package):
@@ -476,7 +546,7 @@ class PackagesDB:
                                 prefer_alt = alternative
                                 prefer_alt_score = 2
                             elif prefer_alt_score < 1 and \
-                                 altdep_state in ["waiting-to-be-tested", "waiting-for-dependency-to-be-tested"]:
+                                    altdep_state in ["waiting-to-be-tested", "waiting-for-dependency-to-be-tested"]:
                                 prefer_alt = alternative
                                 prefer_alt_score = 1
                             elif prefer_alt_score < 0 and altdep_state == "unknown":
@@ -487,28 +557,28 @@ class PackagesDB:
                     if prefer_alt_score >= 0:
                         package.prefer_alt_depends(header, d, prefer_alt)
 
-        deps = package.dependencies()
+        dep_states = [(dep, self.get_best_package_state(dep))
+                      for dep in package.dependencies()]
 
-        for dep in deps:
-            dep_state = self.get_package_state(dep)
+        for dep, dep_state in dep_states:
             if dep_state in self._propagate_error_state:
                 return self._propagate_error_state[dep_state]
 
         testable = True
-        for dep in deps:
-            dep_state = self.get_package_state(dep)
+        for dep, dep_state in dep_states:
             if dep_state not in self._good_states:
                 testable = False
                 break
         if testable:
+            if self._is_successfully_tested(package):
+                return "successfully-tested"
             return "waiting-to-be-tested"
 
         # treat circular-dependencies as testable (for the part of the circle)
         circular_deps = self._get_dependency_cycle(package["Package"])
         if package["Package"] in circular_deps:
             testable = True
-            for dep in deps:
-                dep_state = self.get_package_state(dep)
+            for dep, dep_state in dep_states:
                 if dep in circular_deps:
                     # allow any non-error dep_state on the cycle for testing
                     # (error states are handled by the error propagation above)
@@ -518,24 +588,17 @@ class PackagesDB:
                     testable = False
                     break
             if testable:
+                if self._is_successfully_tested(package):
+                    return "successfully-tested"
                 return "waiting-to-be-tested"
 
-        for dep in deps:
-            dep_state = self.get_package_state(dep)
+        for dep, dep_state in dep_states:
             if dep_state in self._propagate_waiting_state:
                 return self._propagate_waiting_state[dep_state]
 
         return "unknown"
 
-    def _compute_package_states(self):
-        if self._in_state is not None:
-            return
-
-        self._stamp = time.time()
-
-        for subdir in self._all:
-            self._logdb.bulk_load_dir(subdir)
-
+    def _initialize_package_states(self, use_cached_success, check_outdated):
         self._find_all_packages()
 
         self._package_state = {}
@@ -545,7 +608,7 @@ class PackagesDB:
         todo = []
 
         for package_name, package in self._packages.iteritems():
-            state = self._lookup_package_state(package)
+            state = self._lookup_package_state(package, use_cached_success, check_outdated)
             assert state in self._states
             self._package_state[package_name] = state
             if state == "unknown":
@@ -553,8 +616,25 @@ class PackagesDB:
             else:
                 self._in_state[state].append(package_name)
 
+        return todo
+
+    def _compute_package_states(self, use_cached_success=False):
+        if self._in_state is not None:
+            return
+
+        self._stamp = time.time()
+
+        for subdir in self._all:
+            self._logdb.bulk_load_dir(subdir)
+
+        todo = self._initialize_package_states(use_cached_success=use_cached_success, check_outdated=False)
+
         for db in self._dependency_databases:
-            db._compute_package_states()
+            db._compute_package_states(use_cached_success=True)
+
+        if self._dependency_databases:
+            # redo the initialization to properly resolve "outdated" packages after the dependency databases have been initialized
+            todo = self._initialize_package_states(use_cached_success=use_cached_success, check_outdated=True)
 
         while todo:
             package_names = todo
@@ -569,7 +649,7 @@ class PackagesDB:
                     else:
                         self._in_state[state].append(package_name)
                         self._package_state[package_name] = state
-                        done.append(package)
+                        done.append(package_name)
             if not done:
                 # If we didn't do anything this time, we sure aren't going
                 # to do anything the next time either.
@@ -580,10 +660,6 @@ class PackagesDB:
         for state in self._states:
             self._in_state[state].sort()
 
-    def compute_package_states(self):
-        # this must be called with cwd == master-directory
-        self._compute_package_states()
-
     def get_states(self):
         return self._states
 
@@ -618,6 +694,18 @@ class PackagesDB:
                 return self.get_package(providers[0], recurse=recurse, resolve_virtual=False)
         return None
 
+    def get_version(self, name):
+        self._find_all_packages()
+        if name in self._packages:
+            return self._packages[name].version()
+        return None
+
+    def get_test_versions(self, name):
+        self._find_all_packages()
+        if name in self._packages:
+            return self._packages[name].test_versions()
+        return None
+
     def get_providers(self, name, recurse=True):
         self._find_all_packages()
         providers = []
@@ -638,21 +726,7 @@ class PackagesDB:
 
     def get_control_header(self, package_name, header):
         self._find_all_packages()
-        if header == "Source":
-            # binary packages build from the source package with the same name
-            # don't have a Source header, so let's try:
-            try:
-                _source = self._packages[package_name][header]
-                # for binNMU the Source header in Packages files holds the version
-                # too, so we need to chop it of:
-                if " " in _source:
-                    source, version = _source.split(" ")
-                else:
-                    source = _source
-            except:
-                source = self._packages[package_name]["Package"]
-            return source
-        elif header == "Uploaders":
+        if header == "Uploaders":
             # not all (source) packages have an Uploaders header
             uploaders = ""
             try:
@@ -666,6 +740,11 @@ class PackagesDB:
     def get_package_state(self, package_name, resolve_virtual=True, recurse=True):
         self._compute_package_states()
         if package_name in self._package_state:
+            if recurse and self._package_state[package_name] == "outdated":
+                for db in self._dependency_databases:
+                    state = db.get_package_state(package_name, resolve_virtual=resolve_virtual, recurse=False)
+                    if state not in ["does-not-exist", "outdated"]:
+                        return state
             return self._package_state[package_name]
         if package_name in self._virtual_packages:
             if resolve_virtual:
@@ -678,24 +757,65 @@ class PackagesDB:
                 state = db.get_package_state(package_name, resolve_virtual=resolve_virtual, recurse=False)
                 if state != "does-not-exist":
                     return state
-        if package_name in ["ia32-libs-i386", "ia32-libs-gtk-i386"]:
-            # HACK! these are arch=i386 packages needed on amd64
-            return "essential-required"
+        if package_name in self._foreign_packages:
+            return "foreign:%s" % self._foreign_packages[package_name]
         return "does-not-exist"
 
+    def get_best_package_state(self, package_name, resolve_virtual=True, recurse=True):
+        package_state = self.get_package_state(package_name, resolve_virtual=resolve_virtual, recurse=recurse)
+        if package_state in self._good_states:
+            return package_state
+        providers = []
+        if resolve_virtual:
+            providers = self.get_providers(package_name, recurse=recurse)
+        if not providers:
+            return package_state
+        states = [self.get_package_state(name, resolve_virtual=False, recurse=recurse) for name in [package_name] + providers]
+        for state in self._good_states + self._propagate_waiting_state.keys() + self._propagate_error_state.keys():
+            if state in states:
+                return state
+        return package_state
+
+    def _get_package_weight(self, p):
+        # compute the priority of a package that needs testing
+        # result will be used as a reverse sorting key, so higher is earlier
+        waiting_count = self.waiting_count(p["Package"])
+        rdep_chain_len = self.rdep_chain_len(p["Package"])
+
+        if not self._recycle_mode:
+            return (
+                min(rdep_chain_len, waiting_count),
+                    waiting_count,
+            )
+
+        try:
+            statobj = self._logdb.stat(self._recycle, p.name(), p.test_versions())
+            ctime = statobj[stat.ST_CTIME]  # last inode modification = time of linking into recycle/
+            mtime = statobj[stat.ST_MTIME]
+        except OSError:
+            ctime = 0
+            mtime = 0
+
+        return (
+            min(rdep_chain_len, waiting_count),
+                waiting_count,
+                not self._logdb.log_exists(p, [self._ok]),  # prefer problematic logs
+                -ctime / 3600,  # prefer older, at 1 hour granularity to allow randomization
+                -mtime / 3600,  # prefer older, at 1 hour granularity to allow randomization
+        )
+
     def _find_packages_ready_for_testing(self):
         if self._candidates_for_testing is None:
             self._candidates_for_testing = [self.get_package(pn)
-                    for pn in self.get_pkg_names_in_state("waiting-to-be-tested")]
+                                            for pn in self.get_pkg_names_in_state("waiting-to-be-tested")]
             self._candidates_for_testing = [p for p in self._candidates_for_testing
-                    if not self._logdb.log_exists(p, [self._reserved]) or \
-                            self._logdb.log_exists(p, [self._recycle])]
+                                            if not self._logdb.log_exists(p, [self._reserved]) or
+                                            self._logdb.log_exists(p, [self._recycle])]
             if len(self._candidates_for_testing) > 1:
-                self.calc_rrdep_counts()
-                tuples = [(p.waiting_count(), random.random(), p)
-                        for p in self._candidates_for_testing]
-                self._candidates_for_testing = [x[2]
-                        for x in sorted(tuples, reverse = True)]
+                tuples = [(self._get_package_weight(p), random.random(), p)
+                          for p in self._candidates_for_testing]
+                self._candidates_for_testing = [x[-1]
+                                                for x in sorted(tuples, reverse=True)]
         return self._candidates_for_testing[:]
 
     def _remove_unavailable_candidate(self, p):
@@ -707,14 +827,14 @@ class PackagesDB:
             if self._recycle_mode and self._logdb.log_exists(p, [self._recycle]):
                 for vdir in all_but_recycle:
                     if self._logdb.log_exists(p, [vdir]):
-                        self._logdb.remove(vdir, p["Package"], p["Version"])
-                        logging.info("Recycled %s %s %s" % (vdir, p["Package"], p["Version"]))
+                        self._logdb.remove(vdir, p.name(), p.test_versions())
+                        logging.info("Recycled %s %s %s" % (vdir, p.name(), p.test_versions()))
             if self._logdb.log_exists(p, all_but_recycle):
                 self._remove_unavailable_candidate(p)
                 continue
             if self._logdb.log_exists(p, [self._recycle]):
-                self._logdb.remove(self._recycle, p["Package"], p["Version"])
-            if self._logdb.create(self._reserved, p["Package"], p["Version"], ""):
+                self._logdb.remove(self._recycle, p.name(), p.test_versions())
+            if self._logdb.create(self._reserved, p.name(), p.test_versions(), ""):
                 return p
         return None
 
@@ -758,59 +878,102 @@ class PackagesDB:
         else:
             raise LogfileExists(self._evil, package, version)
 
-    def calc_rrdep_counts(self):
-        """Calculate recursive reverse dependency counts for Packages"""
+    def _get_rdep_dict(self):
+        """Return dict of one-level reverse dependencies by package"""
 
-        self._find_all_packages()       # populate _packages
+        if self._rdeps is None:
+
+            self._rdeps = {}
+
+            for pkg_name in self.get_all_package_names():
+                for dep in self.get_package(pkg_name).dependencies():
+                    dep_pkg = self.get_package(dep, recurse=True, resolve_virtual=True)
+
+                    if dep_pkg is not None:
+                        dep = dep_pkg["Package"]
+
+                    if not dep in self._rdeps:
+                        self._rdeps[dep] = set()
+                    self._rdeps[dep].add(pkg_name)
+
+        return self._rdeps
+
+    def _calc_rrdep_pkg_counts(self, pkg):
+
+        pkg_name = pkg['Package']
         self._compute_package_states()  # populate _package_state
+
+        # calc full recursive reverse dependency package set
+        rrdep_set = set()
+        rdeps = self._get_rdep_dict()
+        next_level = set([pkg_name])
+        chain_len = 0
+
+        while next_level:
+            chain_len += 1
+            rrdep_set |= next_level
+            new_pkgs = next_level
+            next_level = set([y for x in new_pkgs if x in rdeps for y in rdeps[x]])
+            next_level -= rrdep_set
+
+        rrdep_set.remove(pkg_name)
+
+        # calculate and set the metrics
+        pkg.rrdep_cnt = len(rrdep_set)
+
         error_states = self.get_error_states()
+        if self._package_state[pkg_name] in error_states:
+            block_list = [x for x in rrdep_set
+                          if self._package_state[x] in error_states]
+            pkg.block_cnt = len(block_list)
+        else:
+            pkg.block_cnt = 0
+
         waiting_states = self.get_waiting_states()
+        if self._package_state[pkg_name] in waiting_states:
+            waiting_list = [x for x in rrdep_set
+                            if self._package_state[x] in waiting_states]
+            pkg.waiting_cnt = len(waiting_list)
+        else:
+            pkg.waiting_cnt = 0
 
-        # create a reverse dependency dictionary.
-        # entries consist of a one-level list of reverse dependency package names,
-        # by package name
-        rdeps = {}
-        for pkg_name in self._packages.keys():
-            # use the Packages dependencies() method for a conservative count
-            for dep in self._packages[pkg_name].dependencies():
-                if dep in rdeps:
-                    rdeps[dep].append(pkg_name)
-                else:
-                    rdeps[dep] = [pkg_name]
+        pkg.rdep_chain_len = chain_len
 
-        def recurse_rdeps(pkg_name, rdeps, rrdep_dict):
-            """ Recurse through the reverse dep arrays to determine the recursive
-                dependency count for a package. rrdep_dict.keys() contains the
-                accumulation of rdeps encountered"""
+    def block_count(self, name):
+        pkg = self.get_package(name)
+        if pkg is None:
+            return -1
+        if pkg.block_cnt is None:
+            self._calc_rrdep_pkg_counts(pkg)
 
-            # are there any rdeps for this package?
-            if pkg_name in rdeps:
-                for rdep in rdeps[pkg_name]:
-                    # break circular dependency loops
-                    if not rdep in rrdep_dict:
-                        rrdep_dict[rdep] = 1
-                        rrdep_dict = recurse_rdeps(rdep, rdeps, rrdep_dict)
+        return pkg.block_cnt
 
-            return rrdep_dict
+    def rrdep_count(self, name):
+        pkg = self.get_package(name)
+        if pkg is None:
+            return -1
+        if pkg.rrdep_cnt is None:
+            self._calc_rrdep_pkg_counts(pkg)
 
-        # calculate all of the rrdeps and block counts
-        for pkg_name in self._packages.keys():
-            rrdep_list = recurse_rdeps(pkg_name, rdeps, {}).keys()
-            self._packages[pkg_name].set_rrdep_count(len(rrdep_list))
+        return pkg.rrdep_cnt
 
-            if self._package_state[pkg_name] in error_states:
-                block_list = [x for x in rrdep_list
-                              if self._package_state[x] in error_states]
-            else:
-                block_list = []
-            self._packages[pkg_name].set_block_count(len(block_list))
+    def waiting_count(self, name):
+        pkg = self.get_package(name)
+        if pkg is None:
+            return -1
+        if pkg.waiting_cnt is None:
+            self._calc_rrdep_pkg_counts(pkg)
 
-            if self._package_state[pkg_name] in waiting_states:
-                waiting_list = [x for x in rrdep_list
-                              if self._package_state[x] in waiting_states]
-            else:
-                waiting_list = []
-            self._packages[pkg_name].set_waiting_count(len(waiting_list))
+        return pkg.waiting_cnt
+
+    def rdep_chain_len(self, name):
+        pkg = self.get_package(name)
+        if pkg is None:
+            return -1
+        if pkg.rdep_chain_len is None:
+            self._calc_rrdep_pkg_counts(pkg)
+
+        return pkg.rdep_chain_len
 
 
 # vi:set et ts=4 sw=4 :
